from typing import Optional



class Compiler:
    def __init__(self, name, version):
        self.name = name
        self.version = version

    def __eq__(self, other: 'Compiler') -> bool:
        return (self.name == other.name) and (self.version == other.version)

    def __str__(self) -> str:
        return f"{self.name}-{self.version}"

    @property
    def aggregates_targets(self) -> bool:
        return self.name == 'vs'

    def name_is(self, name: str) -> bool:
        return self.name == name

    def supports_architecture(architecture: str) -> bool:
        return architecture == 'x86_64'

    @property
    def versioned_name(self) -> str:
        return f"{self.name}-{self.version}"


supported_compilers = (
    Compiler('gcc', '4.2'),
    Compiler('llvm', '19'),
    Compiler('vs', '2017'),
    Compiler('vs', '2022'),
)
names = [ compiler.name for compiler in supported_compilers ]


def from_name_and_version(name: str, version: str) -> Optional[Compiler]:
    for compiler in supported_compilers:
        if compiler.name == name and compiler.version == version:
            return compiler
    return None


def from_versioned_name(versioned_name: Optional[str]) -> Optional[Compiler]:
    if not versioned_name:
        return None
    tokens = versioned_name.split('-')
    if len(tokens) != 2:
        return None
    return from_name_and_version(tokens[0], tokens[1])



import json
import logging
from pathlib import Path
import uuid

from safehouse import services


logger = logging.Logger(__name__)
NULL_UUID = uuid.UUID(int=0)


class User:
    def __init__(self):
        self.ids_by_mode = {
            'live': NULL_UUID,
            'local': NULL_UUID,
            'standalone': NULL_UUID,
        }
        self.safehouse_directory = Path.home() / '.safehouse'
        self.organizations_directory = self.safehouse_directory / 'organizations'
        self.organizations_directory.mkdir(parents=True, exist_ok=True)
        self.load_ids()

    def __str__(self) -> str:
        return '\n\t'.join([f"{k}: {str(v)}" for k, v in self.ids_by_mode.items()])

    def id_for(self, mode: str) -> uuid.UUID:
        return self.ids_by_mode.get(mode, NULL_UUID)

    @property
    def json_file(self) -> Path:
        return self.safehouse_directory / 'user.json'

    def load_ids(self):
        if self.json_file.is_file() is False:
            try:
                self.ids_by_mode['local'] = services.users.get_local_user()['safehouse_id']
                self.save()
            except services.exceptions.ServicesError as e:
                logger.error(f"couldn't retrieve local user: '{e}'")
                return
        with open(self.json_file, 'r', encoding='utf-8') as f:
            self.ids_by_mode = {
                k: uuid.UUID(v)
                for k, v in json.load(f).items()
            }

    def save(self):
        with open(self.json_file, 'w') as f:
            serializable_ids = {
                k: str(v)
                for k, v in self.ids_by_mode.items()
            }
            json.dump(serializable_ids, f, indent=4)

    def set_id_for(self, mode: str, id: uuid.UUID) -> bool:
        if mode not in self.ids_by_mode.keys():
            logger.error(f"unknown mode '{mode}'")
            return False
        self.ids_by_mode[mode] = id
        print(self.ids_by_mode)
        self.save()
        return True



from ..exceptions import CommandError
from . import set_attribute


def run(args: list[str]):
    if len(args) < 1:
        raise CommandError(usage())
    command = args[0]
    if command == 'set':
        set_attribute.run(args[1:])


def usage() -> str:
    return "safehouse users set <attribute.mode=value>"



from typing import Optional

from . import Configuration
from ..builders import Builder


def configuration_to_cmake_target(configuration: Configuration):
    compiler = configuration.compiler
    if compiler.name in ('gcc', 'llvm'):
        return 'Unix Makefiles'
    elif compiler.name_is('vs'):
        if compiler.version == '2017':
            return 'Visual Studio 15 2017 Win64'
        if compiler.version == '2022':
            return 'Visual Studio 17 2022'
    raise Exception(f"cmake doesn't know how to handle compiler '{compiler}'")


def configuration_to_compiler_name(configuration: Configuration) -> str:
    compiler = configuration.compiler
    if compiler.name_is('gcc'):
        return 'GCC with Makefiles'
    elif compiler.name_is('llvm'):
        return 'LLVM with Makefiles'
    elif compiler.name_is('vs'):
        if compiler.version == '2017':
            return 'Visual Studio 2017'
        elif compiler.version == '2022':
            return 'Visual Studio 2022'
    raise Exception(f"cmake doesn't know how to handle compiler '{compiler}'")


def configuration_to_extra_arguments(configuration: Configuration) -> str:
    compiler = configuration.compiler
    if compiler.name_is('gcc'):
        return "-DCMAKE_C_COMPILER=gcc -DCMAKE_CXX_COMPILER=g++ -DCMAKE_BUILD_TYPE=Debug"
    elif compiler.name_is('llvm'):
        return '-DCMAKE_C_COMPILER=clang-19 -DCMAKE_CXX_COMPILER=clang++-19 -DCMAKE_BUILD_TYPE=Debug'
    elif compiler.name_is('vs'):
        if compiler.version == '2022':
            return '-A x64'
    return ''


def configuration_to_toolchain(configuration: Configuration) -> Optional[str]:
    return None

def generate_build_from_configuration(configuration: Configuration, builder: Builder):
    generate_build(
        archName='x86 64-bit',
        builder=builder,
        cmakeTarget=configuration_to_cmake_target(configuration),
        compilerName=configuration_to_compiler_name(configuration),
        extraArgs=configuration_to_extra_arguments(configuration),
        noGraphics=False,
        outDir=configuration.output_directory,
        buildNameDir=configuration.build_directory,
        pathToSource=configuration.source_directory,
        platformName=configuration.platform.name.capitalize(),
        toolchain=configuration_to_toolchain(configuration),
        unityBuild=False,
    )


def generate_build(
        *,
        archName: str,
        builder: Builder,
        cmakeTarget: str,
        compilerName: str,
        extraArgs: Optional[str]=None,
        noGraphics: bool,
        outDir: str,
        buildNameDir: str,
        pathToSource: str,
        platformName: str,
        toolchain: Optional[str]=None,
        unityBuild: bool,
):
    # determine whether or not we need graphics.
    noGraphicsStr = "-DNO_SYSTEM_GRAPHICS=0"
    if noGraphics:
        noGraphicsStr = "-DNO_SYSTEM_GRAPHICS=1"
        #outDir += "_NO_GFX"

    # print information about the build system we're preparing to generate.
    print( "Generating build system..." )
    print( outDir )
    print( "Target platform:  {}".format( platformName ) )
    print( "Target processor: {}".format( archName ) )
    print( "Target compiler:  {}".format( compilerName ) )

    # set the current working directory to the output directory.
    fullOutDir = outDir

    # run CMake in the output folder.
    unityBuildStr = "-DUNITY_BUILD=0"
    if unityBuild:
        unityBuildStr = "-DUNITY_BUILD=1"

    # security option string.
    secureStr = "-DSECURE_BUILD=0"
    #if options.secure:
    #    secureStr = "-DSECURE_BUILD=1"

    if toolchain is not None:
        cmakeExecStr = f'cmake -DBUILD_DIRECTORY={outDir} {pathToSource} -G"{cmakeTarget}" ' \
                       f'-T"{toolchain}" --no-warn-unused-cli -DBUILD_CONFIG_DIR="{buildNameDir}" ' \
                       f'{unityBuildStr} {noGraphicsStr} {secureStr}'
    else:
        cmakeExecStr = f'cmake -DBUILD_DIRECTORY={outDir} {pathToSource} -G"{cmakeTarget}" ' \
                       f'--no-warn-unused-cli -DBUILD_CONFIG_DIR="{buildNameDir}" '\
                       f'{unityBuildStr} {noGraphicsStr} {secureStr}'

    if extraArgs is not None:
        cmakeExecStr = '{0} {1}'.format( cmakeExecStr, extraArgs )

    builder.build(fullOutDir, cmakeExecStr)

    print(f"cmakeExecStr=='{cmakeExecStr}'")
    #prevDir = os.getcwd()
    ## change to the output folder, making it if needed.
    #if not os.path.exists( fullOutDir ):
    #    os.makedirs( fullOutDir )
    #os.chdir( fullOutDir )
#
#    p = subprocess.Popen(cmakeExecStr, stdout=subprocess.PIPE, shell=True, universal_newlines=True)
#    if p:
#        (output, err) = p.communicate()
#        print( output )
#
#    # restore the previous working directory.
#    os.chdir( prevDir )



import datetime
import json
import logging

from safehouse.services import client
from safehouse.types import JsonBlob


logger = logging.Logger(__name__)


def get_event_types(organization_name: str, project_name: str) -> list[JsonBlob]:
    response = client.get(
        endpoint='/events/types/registry/',
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-ORGANIZATION": organization_name,
            "X-PROJECT": project_name,
        }
    )
    return response.json()


def register_event_types(
    *,
    event_types: list[JsonBlob],
) -> list[JsonBlob]:
    response = client.post(
        data=json.dumps(event_types),
        endpoint='/events/types/registry/',
    )
    return response.json()


def send_event(
    *,
    attributes: JsonBlob,
    name: str,
    origin: str,
    sent_at: datetime.datetime,
    version: int,
) -> bool:
    try:
        client.post(
            data=json.dumps(
                {
                    'attributes': attributes,
                    'name': name,
                    'origin': origin,
                    'sent_at': sent_at.isoformat(),
                    'version': version,
                }
            ),
            endpoint='/events/',
        )
    except client.ClientError as e:
        logger.error(f"couldn't send event type '{name}.{version}': {e}")
        return False
    return True



import os
from typing import List, Optional


class Project:
    def __init__(
            self,
            *,
            dependencies: Optional[List['Project']]=[],
            name,
            relative_filepath: Optional[str]=None,
            source_directory: Optional[str]=None,
    ):
        # TODO - make sure there aren't recursive dependencies
        self.name = name
        self.dependencies = dependencies
        self.relative_filepath = relative_filepath
        self.source_directory = source_directory

    @property
    def is_installed(self, build_prefix: str) -> bool:
        return False

    def missing_dependencies(self, build_prefix: str) -> List['Project']:
        '''
        Return dependent projects that haven't been built under build_prefix
        '''
        return []

    @property
    def relative_build_path(self) -> str:
        '''
        The build path for this project relative to SHMAKE_SOURCE
        '''
        if self.relative_filepath:
            return os.path.join(self.relative_filepath, self.name)
        return self.name

    @property
    def relative_source_path(self) -> str:
        '''
        The source code path for this project relative to SHMAKE_SOURCE
        '''
        path = self.name
        if self.relative_filepath:
            path = os.path.join(self.relative_filepath, path)
        if self.source_directory:
            path = os.path.join(path, self.source_directory)
        return path
    
    @property
    def source_available(self, source_prefix: str) -> bool:
        '''
        Return True if the source code for this project exists where expected, False if not.
        '''
        return False


external_project = Project(name='external', relative_filepath='shared')
shared_project = Project(name='shared', dependencies=[external_project])
engine_project = Project(
    name='ion',
    dependencies=[external_project, shared_project],
    source_directory='code',
)


supported_projects = (
    external_project,
    shared_project,
    engine_project,
)


names = [ project.name for project in supported_projects ]


def from_name(project_name: str) -> Project:
    for project in supported_projects:
        if project.name == project_name:
            return project 
    return None



from importlib.resources import files
from typing import List

from python_on_whales import DockerClient

from .exceptions import DockerComposeError
from .types import Container


docker_compose_file = str(files('safehouse.data').joinpath('docker-compose.yaml'))
docker = DockerClient(compose_files=[docker_compose_file])
if not docker.compose.is_installed():
    raise DockerComposeError("Please install docker (https://docs.docker.com/get-started/get-docker/)")

docker.run.__doc__

def down():
    docker.compose.down()


def is_running(service_name) -> bool:
    return service_name in [container.name for container in ps()]
 

def ps() -> List[Container]:
    return docker.ps()
 

def up(detach: bool=False):
    docker.compose.up(detach=detach)



class Target:
    def __init__(self, name):
        self.name = name

    def __str__(self) -> str:
        return self.name


supported_targets = (
    Target('debug'),
    Target('minsizerel'),
    Target('release'),
    Target('relwithdb'),
)

names = [ target.name for target in supported_targets]


def from_name(name: str) -> Target:
    for target in supported_targets:
        if target.name == name:
            return target 
    return None



import datetime
import logging
import uuid

from .exceptions import FieldError
from safehouse.services import events as events_service
from safehouse.types import JsonBlob


FIELD_TYPES = int|float|str # Datetime is handled as a special case
logger = logging.Logger(__name__)


class FieldType:
    def __init__(self, data_type: FIELD_TYPES):
        self.data_type: FIELD_TYPES = data_type

    def cast(self, value: object) -> FIELD_TYPES:
        try:
            return self.data_type(value) # ty: ignore
        except ValueError:
            raise FieldError(f"couldn't convert '{value}' to data type {self.data_type}")


class DatetimeFieldType:
    def __init__(self):
        pass

    def cast(self, value: datetime.datetime) -> str:
        try:
            return value.isoformat()
        except Exception:
            raise FieldError(f"couldn't convert '{value}' to a stringified datetime") 


class UUIDFieldType:
    def __init__(self):
        pass

    def cast(self, value: uuid.UUID) -> str:
        try:
            return str(value)
        except Exception:
            raise FieldError(f"couldn't convert '{value}' to a UUID")


DATA_TYPE_NAME_TO_FIELD_TYPE = {
    'datetime': DatetimeFieldType(),
    'int': FieldType(int), # ty: ignore
    'float': FieldType(float), # ty: ignore
    'str': FieldType(str), # ty: ignore
    'uuid': UUIDFieldType(),
}


class EventType:
    def __init__(
        self,
        *,
        fields: dict[str, str]={},
        name: str,
        organization_name: str,
        project_name: str,
        services_url: str,
        version: int,
    ):
        self.fields = {}
        for field_name, field_type_name in fields.items():
            field_type = DATA_TYPE_NAME_TO_FIELD_TYPE.get(field_type_name)
            if not field_type:
                raise FieldError(f"unknown field type '{field_type_name}'")
            self.fields[field_name] = field_type
        self.name = name
        self.organization_name= organization_name
        self.project_name = project_name
        self.services_url = services_url
        self.version = version

    def __str_(self) -> str:
        return self.name


class Event:
    def __init__(
            self,
            *,
            event_type: EventType,
            origin: str,
    ):
        self.event_type = event_type
        self.origin = origin

    def __str__(self) -> str:
        return f"{self.event_type} event"

    def format_attributes(self, attributes: JsonBlob) -> JsonBlob:
        formatted_attributes = {}
        for name, value in attributes.items():
            field_type = self.event_type.fields.get(name)
            if not field_type:
                raise Exception(f"attempting to send unknown field '{name}={attributes[name]}'")
            formatted_attributes[name] = field_type.cast(value)
        return formatted_attributes

    def send(self, *args, **kwargs) -> bool:
        '''
        send event with supplied arguments; returns True if sent and False otherwise.
        '''
        try:
            event_attributes = self.format_attributes(kwargs)
            return events_service.send_event(
                attributes=event_attributes,
                name=self.event_type.name,
                origin=self.origin,
                sent_at=datetime.datetime.now(datetime.timezone.utc),
                version=self.event_type.version,
            )
        except FieldError as e:
            logger.error(f"couldn't send {self}: {e}")
            return False


class UnknownEvent:
    def __init__(
            self,
            *,
            event_type_name: str,
            origin: str,
    ):
        self.event_type_name = event_type_name
        self.origin = origin

    def send(self, *args, **kwargs) -> bool:
        logger.error(f"unknown event type: '{self.event_type_name}'")
        return False



import logging
import os

from . import command
from .. import config


logger = logging.getLogger()


class Build(command.Command):
    def __init__(self):
        super().__init__()
        self.configuration = config.configuration

    def __str__(self) -> str:
        return f"configure {self.configuration}"

    def configureFromArguments(self):
        super().configureFromArguments()

    @property
    def description(self) -> str:
        return 'build the current directory'

    def initialize_options(self):
        pass

    @property
    def name(self) -> str:
        return 'build'

    def run(self) -> bool:
        print(self)
        os.system('make')
        return True



import logging
import platform
from typing import Optional

from . import command
from .. import builders, configurations, projects
from ..configurations import cmake, compilers, platforms, targets


logger = logging.getLogger()


class Configure(command.Command):
    def __init__(self):
        super().__init__()
        self.configuration: Optional[configurations.Configuration] = None

    def __str__(self) -> str:
        return f"configure {self.configuration}"

    def configureFromArguments(self):
        self.parser.add_argument(
            '--compiler',
            dest='compiler',
            help='Use the specified compiler-version.',
            required=False,
        )
        self.parser.add_argument(
            "--no_graphics",
            "--nogfx",
            default=False,
            dest="no_graphics",
            help="Build with no graphics functionality (may not be supported by all builds).",
            required=False,
        )
        self.parser.add_argument(
            '--project',
            choices=projects.names,
            dest='project',
            help='The project to configure.',
        )
        self.parser.add_argument(
            '--secure',
            '--s',
            default=False,
            dest='secure',
            help='Generate a secure build.',
            required=False,
        )
        self.parser.add_argument(
            '--target',
            dest='target',
            help='Build target configuration.',
            required=False,
        )
        self.parser.add_argument(
            "--unity_build",
            "--unity",
            default=False,
            dest="unity_build",
            help="Generate a unity build.",
            required=False,
        )
        super().configureFromArguments()

    @property
    def description(self) -> str:
        return 'create a build configuration'

    def initialize_options(self):
        options = self.options
        if not options.project:
            raise Exception(f"no project specified; options are {','.join(projects.names)}")
        self.platform = platforms.from_name(options.platform)
        if not self.platform:
            raise Exception(f"platform {options.platform} is not supported")
        if options.platform != platform.system().lower() and options.platform != 'linux':
            raise Exception("cross-compiling not yet supported")
        self.compiler = compilers.from_versioned_name(options.compiler) if options.compiler else self.platform.default_compiler
        if not self.platform.supports_compiler(self.compiler):
            raise Exception(f"compiler {self.compiler} unsupported on {self.platform}")
        target_name = options.target if options.target else 'debug'
        self.target = targets.from_name(target_name)
        if not self.target:
            raise Exception(f"unknown target '{target_name}'")

    @property
    def name(self) -> str:
        return 'configure'

    def run(self) -> bool:
        # If configuration is for linux and host os is not linux run in linux builder container
        self.configuration = configurations.from_options(self.options)
        print(self)
        # TIM - TODO: handle cmake failures
        cmake.generate_build_from_configuration(self.configuration, builders.local)
        self.configuration.save()
        return True



import json
import logging
import requests
from typing import List, Optional

import safehouse


logger = logging.Logger(__name__)


class ClientError(Exception):
    def __init__(self, detail):
        super().__init__(detail)


def get(
        *,
        acceptable_status_codes: Optional[List[int]] = [200],
        endpoint: str,
        headers: dict[str,str] | None=None,
) -> requests.Response:
        if headers is None:
            headers = _headers()
        response = requests.get(
            url=services_url() + endpoint,
            headers=headers,
        )
        validate_response(response, acceptable_status_codes=acceptable_status_codes)
        return response


def post(
        *,
        acceptable_status_codes: Optional[List[int]] = [200],
        data: Optional[str]='',
        endpoint: str,
        headers: dict[str,str] | None=None,
) -> requests.Response:
        if headers is None:
            headers = _headers()
        response = requests.post(
            url=services_url() + endpoint,
            headers=headers,
            data=data,
        )
        validate_response(response, acceptable_status_codes=acceptable_status_codes)
        return response


def services_url():
    if safehouse.active_project:
        return safehouse.active_project.services_url
    return safehouse.mode.services_url()


def validate_response(
        response: requests.Response,
        acceptable_status_codes: Optional[List[int]] = [200],
):
    if acceptable_status_codes and response.status_code not in acceptable_status_codes:
        error_string = str(response)
        error_detail = error_string
        try:
            error_string = response.json()
            if type(error_string) is list:
                error_detail = ', '.join(error_string)
            else:
                error_detail = error_string['detail']
        except json.JSONDecodeError:
            pass
        raise ClientError(error_detail)


def _headers():
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
    }
    if safehouse.active_project:
        if safehouse.user:
            headers['X-USER-ID'] = str(safehouse.user.id_for(safehouse.active_project.mode))
        headers['X-ORGANIZATION'] = safehouse.active_project.organization_name
        headers['X-PROJECT'] = safehouse.active_project.name
    return headers



import logging

from . import (
    events as events,
    mode as mode,
    projects as projects,
    services as services,
    users as users,
    utils as utils,
)


active_project = None
logger = logging.Logger(__name__)
runtime_mode = "local"
user = users.User()


projects.init(user.safehouse_directory, user.organizations_directory)


def activate_project(
    *,
    name: str,
    org_name: str,
    runtime_mode: str,
    user: users.User | None = None,
) -> projects.Project:
    global active_project
    active_project = projects.register_project(
        mode=runtime_mode,
        name=name,
        org_name=org_name,
        user=user,
    )
    try:
        services.organizations.register_active_project()
    except services.exceptions.ServicesError as e:
        logger.error(f"couldn't register {active_project} with safehouse services: '{e}'")
    return active_project



from safehouse.services import events
from ..exceptions import CommandError


def run(args: list[str]):
    if len(args) != 1:
        raise CommandError(usage())
    tokens = args[0].split('.')
    if len(tokens) != 2:
        raise CommandError(usage())
    organization, project = tokens
    events_list = events.get_event_types(organization, project)
    events_list = sorted(events_list, key=lambda e: e['name'])
    print(f"Registered events for {organization}.{project}:")
    for e in events_list:
        name = e['name']
        attributes: dict[str, str] = e['attributes'] # ty: ignore
        s = f"\t{name}: "
        s += ', '.join([f"{k}:{v}" for k, v in attributes.items()])
        print(s)


def usage() -> str:
    return "usage: safehouse projects list_events <organization.project>"



import logging

import safehouse
from safehouse import assertions
from safehouse.services import events as events_service
from safehouse.types import JsonBlob

from .event import Event, EventType, UnknownEvent


logger = logging.Logger(__name__)


class EventManager:
    def __init__(self, event_type_definitions: list[JsonBlob], origin: str=''):
        self.event_types: dict[str, EventType] = {}
        self.event_type_definitions = event_type_definitions
        self.origin = origin

    def __getattr__(self, name):
        event_type = self.event_types.get(name)
        if event_type:
            return Event(event_type=event_type, origin=self.origin)
        return UnknownEvent(event_type_name=name, origin=self.origin)

    def register_event_types(self):
        assertions.active_project()
        event_types_data: list[JsonBlob] = []
        event_types_data= events_service.register_event_types(
            event_types=self.event_type_definitions,
        )
        self.event_types = {
            str(e['name']): EventType(
                fields=e['attributes'], # ty: ignore
                name=e['name'], #ty: ignore
                organization_name=self.organization_name,
                project_name=self.project_name,
                services_url=self.services_url,
                version=e['version'], # ty: ignore
            ) # ty: ignore
            for e in event_types_data
        }
        if not self.event_types:
            logger.warning(f"There are no event types available for use in {safehouse.active_project}")



from pathlib import Path

import safehouse
from safehouse import users, utils


from .projects import (
    from_safehouse_dir as from_safehouse_dir,
    from_settings as from_settings,
    Project as Project,
)
from . import exceptions as exceptions


_projects_registry: dict[str, dict[str, Project]] = {}


def current_development_project() -> Project | None:
    for p in list_projects():
        if p.in_development:
            return p
    return None

        
def init(user_safehouse_directory, user_organizations_directory):
    development_safehouse_directory = utils.find_directory_in_parent_dirs('.safehouse')
    if development_safehouse_directory and development_safehouse_directory != user_safehouse_directory:
        load_project(from_safehouse_dir(development_safehouse_directory, in_development=True))
    load_projects(user_organizations_directory)


def load_project(project: Project):
    if project.organization_name not in _projects_registry:
        _projects_registry[project.organization_name] = { project.name: project }
    else:
        _projects_registry[project.organization_name][project.name] = project


def load_projects(organizations_root: Path):
    organization_dirs = [item for item in organizations_root.iterdir() if item.is_dir()]
    for organization_dir in organization_dirs:
        projects_dirs = [item for item in organization_dir.iterdir() if item.is_dir()]
        for project_dir in projects_dirs:
            if not registered_project(organization_dir.name, project_dir.name):
                load_project(from_safehouse_dir(project_dir))


def list_projects() -> list[Project]:
    projects_list = []
    for org_projects in _projects_registry.values():
        projects_list += org_projects.values()
    return projects_list


def register_project(
    *,
    mode: str,
    name: str,
    org_name: str,
    user: users.User | None = None,
) -> Project:
    project = registered_project(org_name, name)
    if project:
        return project
    if user is None:
        user = safehouse.user
    filepath = user.organizations_directory / org_name / name
    filepath.mkdir(parents=True, exist_ok=True)
    project = from_settings(filepath=filepath, mode=mode, org_name=org_name, name=name)
    project.save()
    return project


def registered_project(org_name: str, name: str) -> Project | None:
    org_dict = _projects_registry.get(org_name)
    if org_dict:
        return org_dict.get(name)
    return None



#!/usr/bin/env python3

from importlib import metadata
import logging
from pathlib import Path
import sys

from . import commands, config


logger = logging.getLogger()


COMMANDS = { 
    'configure': commands.Configure,
    'build': commands.Build,
}


def run():
    print(f"shmake-{metadata.version('safehouse')}")
    logging.basicConfig(
        filename=Path('/tmp/shmake.log'),
        level=logging.INFO,
    )
    command_class = None
    if len(sys.argv) > 1:
        command_class_from_args = COMMANDS.get(sys.argv[1], command_class)
        if command_class_from_args:
            command_class = command_class_from_args
    if not command_class:
        command_class = commands.Build if config.configuration else commands.Configure
    try:
        command: commands.Command = command_class()
        command.configureFromArguments()
        command.run()
    except Exception as e:
        print(e)
        print("run shmake --help for more information")
        raise(e)

if __name__ == "__main__":
    run()



from abc import ABC, abstractmethod
from argparse import ArgumentParser
import platform

from ..configurations import platforms


class Command(ABC):
    def __init__(self):
        self.parser = ArgumentParser(prog=f"schmake {self.name}")

    @abstractmethod
    def configureFromArguments(self):
        self.parser.add_argument(
            'command',
            default=self.name,
            nargs='?',
            help='command to run',
        )
        # TIM - move to configure
        self.parser.add_argument(
            '--platform',
            default=platform.system().lower(),
            choices=platforms.names,
            dest='platform',
            required=False,
        )
        self.options = self.parser.parse_args()
        self.initialize_options()

    @property
    @abstractmethod
    def description(self) -> str:
        return 'abstract description'

    @property
    @abstractmethod
    def name(self) -> str:
        return 'abstract name'

    def print_usage(self):
        self.parser.print_usage()

    @abstractmethod
    def run(self) -> bool:
        return False

    @abstractmethod
    def initialize_options(self):
        pass

    def __str__(self) -> str:
        return f"{self.name}"



import safehouse


def run(args: list[str]):
    mode = safehouse.runtime_mode
    print(f"mode: {mode}, user_id: {safehouse.user.id_for(mode)}")
    projects = safehouse.projects.list_projects()
    if projects:
        print("registered projects:")
        for p in projects:
            s = f"\t{p}"
            if p.in_development:
                s += " (in development)"
            print(s)



import logging
import uuid

import safehouse
from safehouse.users import NULL_UUID
from ..exceptions import CommandError


logger = logging.Logger(__name__)


def run(args: list[str]):
    if len(args) != 1:
        raise CommandError(usage())
    command_tokens = args[0].split('=')
    if len(command_tokens) != 2:
        raise CommandError(usage())
    value = command_tokens[-1]
    attribute_tokens = command_tokens[0].split('.')
    if len(attribute_tokens) != 2:
        raise CommandError(usage())
    attribute, mode = attribute_tokens
    if attribute != 'id':
        raise CommandError(usage())
    try:
        if value == 'null':
            id = NULL_UUID
        else:
            id = uuid.UUID(value)
    except ValueError as e:
        raise CommandError(str(e))
    if (id != NULL_UUID) and (id.version != 4):
        raise CommandError(f"'{value}' is not a valid uuid4")
    if not safehouse.user.set_id_for(mode, id):
        logger.error(f"failed to set user id for mode {mode} to {id}")


def usage() -> str:
    return "safehouse users set id.<mode>=<value|null>"



import logging

from safehouse.services import client
from safehouse.types import JsonBlob


logger = logging.Logger(__name__)


def get_local_user() -> JsonBlob:
    response = client.get(
        endpoint='/users/local/',
    )
    return response.json()



import json
import logging

import safehouse
from safehouse.services import client
from safehouse.types import JsonBlob


logger = logging.Logger(__name__)


def register_active_project() -> JsonBlob:
    assert safehouse.active_project is not None
    response = client.post(
        data=json.dumps({'name': safehouse.active_project.name}),
        endpoint='/organizations/projects/register/',
    )
    return response.json()
    


from typing import List

from . import compilers

# How to easily articulate the supported versions of compilers per platform?


class Platform:
    def __init__(self, name: str):
        self.name = name
        self.compilers: List[compilers.Compiler] = []
        self.default_compiler = None

    def add_compiler(self, compiler: compilers.Compiler, *, default: bool=False):
        self.compilers.append(compiler)
        if default:
            self.default_compiler = compiler
        
    def supports_compiler(self, compiler: compilers.Compiler) -> bool:
        return compiler in self.compilers

    def __str__(self) -> str:
        return self.name


linux_platform = Platform('linux')
osx_platform = Platform('darwin')
windows_platform = Platform('windows')


supported_platforms = (
    linux_platform,
    osx_platform,
    windows_platform,
)

names = [ platform.name for platform in supported_platforms ]


def from_name(platform_name: str) -> Platform:
    for platform in supported_platforms:
        if platform.name == platform_name:
            return platform
    return None



import logging
from pathlib import Path
from typing import List

from safehouse import projects
from ..exceptions import CommandError


logger = logging.getLogger()


def run(args: List[str]) -> projects.Project:
    if projects.current_development_project():
        raise CommandError(f"project already exists: {projects.current_development_project()}")
    if len(args) != 1:
        raise CommandError(usage())
    tokens = args[0].split('.')
    if len(tokens) != 2:
        raise CommandError(usage())
    organization, project = tokens
    project_dir: Path = Path.cwd()
    config_dir = project_dir / '.safehouse'
    try:
        project = projects.from_settings(
            filepath=config_dir,
            name=project,
            org_name=organization,
            mode='local',
        )
    except projects.exceptions.ProjectError as e:
        raise CommandError(str(e))
    Path.mkdir(config_dir, exist_ok=True)
    project.save()
    logger.info(f"succesfully initialized {project}")
    return project


def usage() -> str:
    return "usage: safehouse init <organization.project>"



#!/usr/bin/env python
import logging
import sys

from safehouse.apps.console import commands
from safehouse.apps.console.commands.types import Runnable

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


supported_commands: dict[str, Runnable] = {
    'orgs': commands.orgs.run,
    'projects': commands.projects.run,
    'services': commands.services.run,
    'status': commands.status.run,
    'users': commands.users.run,
    'version': commands.version.run,
}

def run():
    #print(f"safehouse-{metadata.version('safehouse')}")
    args = [arg.lower() for arg in sys.argv[1:]]
    if len(args) > 0:
        command_name: str = args[0]
        try:
            command = supported_commands[command_name]
        except KeyError:
            logger.error(f"'safehouse {command_name}' not supported")
            logger.info(usage())
            sys.exit(-1)
        try:
            command(args[1:])
        except commands.exceptions.CommandError as e:
            logger.error(e)
            sys.exit(-1)
        except Exception as e:
            logger.exception(e)
            sys.exit(-1)
    else:
        print(usage())
        sys.exit(0)


def usage() -> str:
    return f"usage: safehouse <{'|'.join(supported_commands.keys())}>"

if __name__ == "__main__":
    run()



from typing import Optional
from pathlib import Path


def find_directory_in_parent_dirs(dirname, start_path=None) -> Optional[Path]:
    """
    Recursively checks for a file in the current directory and its parent directories.

    Args:
        dirname (str): The name of the file to search for.
        start_path (str or Path, optional): The starting directory for the search.
                                            If None, the current working directory is used.

    Returns:
        Path or None: The absolute path to the file if found, otherwise None.
    """
    if start_path is None:
        current_dir = Path.cwd()
    else:
        current_dir = Path(start_path).resolve()

    while True:
        dir_path = current_dir / dirname 
        if dir_path.is_dir():
            return dir_path
        
        # If we've reached the root directory and the file isn't found, stop
        if current_dir == current_dir.parent:
            return None
        
        current_dir = current_dir.parent


def find_file_in_parent_dirs(filename, start_path=None) -> Optional[Path]:
    """
    Recursively checks for a file in the current directory and its parent directories.

    Args:
        filename (str): The name of the file to search for.
        start_path (str or Path, optional): The starting directory for the search.
                                            If None, the current working directory is used.

    Returns:
        Path or None: The absolute path to the file if found, otherwise None.
    """
    if start_path is None:
        current_dir = Path.cwd()
    else:
        current_dir = Path(start_path).resolve()

    while True:
        file_path = current_dir / filename
        if file_path.is_file():
            return file_path
        
        # Move up to the parent directory
        current_dir = current_dir.parent
        
        # If we've reached the root directory and the file isn't found, stop
        if current_dir == current_dir:
            return None
 


from . import (
    exceptions as exceptions,
    orgs as orgs,
    projects as projects,
    services as services,
    status as status,
    types as types,
    users as users,
    version as version,
)



import argparse
import json
import logging
import os
from pathlib import Path
from typing import Optional

from safehouse import config
from safehouse.apps.shmake import projects
from . import compilers, platforms, targets
from .compilers import Compiler
from .platforms import Platform
from .targets import Target


logger = logging.getLogger()


class Configuration:
    def __init__(
            self,
            *,
            compiler: Compiler,
            platform: Platform,
            project: projects.Project,
            target: Target,
    ):
        self.architecture = 'x86_68'
        self.compiler = compiler
        self.platform = platform
        self.project = project
        self.target = target
        self.validate()

    @property
    def output_directory(self) -> str:
        '''
        Returns root directory for the build artifacts.
        '''
        if self.compiler.aggregates_targets:
            return os.path.join(
                self.trunk_directory,
                f"builds/{self.platform.name}/",
                f"{self.compiler.versioned_name}/",
                self.project.relative_build_path,
            )
        return os.path.join(
            self.trunk_directory,
            f"builds/{self.platform.name}/",
            f"{self.compiler.versioned_name}/",
            f"{self.target}/",
            self.project.relative_build_path,
        )
    
    @property
    def build_directory(self) -> Path:
        '''
        Returns the root directory for the build.
        '''
        path = Path(self.trunk_directory) / f"builds/{self.platform.name}" / self.compiler.versioned_name
        if not self.compiler.aggregates_targets:
            path = path / self.target.name
        path = path / self.project.relative_source_path
        return path
    
    def save(self):
        data = {
            'architecture': self.architecture,
            'compiler': f"{self.compiler.name}-{self.compiler.version}",
            'platform': self.platform.name,
            'project': self.project.name,
            'target': self.target.name,
        }
        try:
            build_configuration_directory = self.build_directory / '.shmake'
            print(f"build_configuration_directory=={build_configuration_directory}")
            with open(build_configuration_directory, 'w') as f:
                json.dump(data, f, indent=4)
        except IOError as e:
            logger.error(e)

    @property
    def source_directory(self) -> Path:
        return Path(self.trunk_directory) / self.project.relative_source_path

    def validate(self):
        if not config.project:
            raise Exception("no trunk directory found, please run `safehouse init` in your trunk directory")
        self.config_directory = config.project.filepath
        self.trunk_directory = self.config_directory.parent
        print(f"trunk directory == {self.trunk_directory}")
        # make sure compiler, platform and target play well together and if so generate the build for this configuration. if not raise exception
        if not self.platform.supports_compiler(self.compiler):
            raise Exception(f"compiler {self.compiler} unsupported on {self.platform}")
        #if not self.compiler.supports_target(self.target):
        #    raise Exception(f"compiler {self.compiler} doesn't support target {self.target}")
        
    def __str__(self) -> str:
        return f"{self.architecture} {self.target} build on {self.platform} using " \
               f"{self.compiler} from {self.trunk_directory}"
        

# =============================================================================
# Define supported compiler versions
# =============================================================================
gcc_4_2 = compilers.from_name_and_version('gcc', '4.2')
llvm_19 = compilers.from_name_and_version('llvm', '19')
vs_2017 = compilers.from_name_and_version('vs', '2017')
vs_2022 = compilers.from_name_and_version('vs', '2022')
# =============================================================================


# =============================================================================
# Define supported platforms and the compilers they support
# =============================================================================
linux_platform = platforms.from_name('linux')
linux_platform.add_compiler(gcc_4_2)
linux_platform.add_compiler(llvm_19, default=True)

osx_platform = platforms.from_name('darwin')
osx_platform.add_compiler(gcc_4_2, default=True)

windows_platform = platforms.from_name('windows')
windows_platform.add_compiler(vs_2017)
windows_platform.add_compiler(vs_2022, default=True)
# =============================================================================


def configuration_for(
        *,
        compiler_name: Optional[str]=None,
        platform_name: Optional[str]=None,
        project_name: str,
        target_name: Optional[str]=None,
) -> Configuration:
    configured_project = projects.from_name(project_name)
    if not project_name:
        raise Exception(f"project {project_name} is not supported")
    #if platform_name != platform.system().lower():
    #    raise Exception("cross-compiling not yet supported")
    configured_platform = platforms.from_name(platform_name)
    if not configured_platform:
        raise Exception(f"platform {platform_name} is not supported")
    compiler = compilers.from_versioned_name(compiler_name)
    if compiler_name and not compiler:
        raise Exception(f"unknown compiler'{compiler_name}")
    compiler = compiler if compiler else configured_platform.default_compiler
    target_name = target_name if target_name else 'debug'
    target = targets.from_name(target_name)
    if not target:
        raise Exception(f"unknown target '{target_name}'")
    return Configuration(
        compiler=compiler,
        platform=configured_platform,
        project=configured_project,
        target=target,
    )


def from_options(options: argparse.Namespace) -> Configuration:
    return configuration_for(
        compiler_name=options.compiler,
        platform_name=options.platform,
        project_name=options.project,
        target_name=options.target,
    )


# Move to build?
def from_file(filepath: Path) -> Configuration:
    try:
        with open(filepath, 'r') as f:
            data = json.load(f)
        compiler = data.get('compiler')
        platform = data.get('platform')
        project = data.get('project')
        target = data.get('target')
        return configuration_for(
            compiler_name=compiler,
            platform_name=platform,
            project_name=project,
            target_name=target,
        )
    except FileNotFoundError:
        logger.error(f"file '{filepath} not found")
    except json.JSONDecodeError:
        logger.error(f"couldn't decode JSON from '{filepath}")
    


from ..exceptions import CommandError
from . import (
    init as init,
    list_events as list_events,
)


def run(args: list[str]):
    if len(args) != 2:
        raise CommandError(usage())
    command = args[0]
    if command == 'init':
        init.run(args[1:])
    elif command == 'list_events':
        list_events.run(args[1:])
    else:
        raise CommandError(f"unsupported projects command '{command}'")


def usage() -> str:
    return "safehouse projects init <organization.project>"
 


from abc import ABC, abstractmethod
import os
from pathlib import Path
import subprocess


class Builder(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def build(self, build_dir: Path, command: str):
        pass

    def connect(self):
        '''
        Connect to the guest os for the builder (if there is one)
        '''
        pass

class LocalBuilder(Builder):
    def __init__(self):
        super().__init__()

    def build(self, build_dir: Path, command: str):
        prev_dir = os.getcwd()
        if not os.path.exists(build_dir):
            os.makedirs(build_dir)
        os.chdir(build_dir)
        p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True, universal_newlines=True)
        if p:
            output, _ = p.communicate()
            print(output)
        os.chdir(prev_dir)


class LinuxBuilder(Builder):
    def __init__(self):
        super().__init__()

    def build(self, build_dir: Path, command: str):
        '''
        download linux builder image if it doesn't locally
        run builder in container:
        - mount the build dir as /src
        - command to run is command passed in to function
        '''
        pass

# NOTE: cd into directory invoked from for `shmake builder connect`
# When configuring or running a linux build from a non-linux host, prompt to start container, if not fail.
# - start container with tail -f /dev/null, then issue it the build command
local = LocalBuilder()



from safehouse import utils

from . import configurations

configuration = None
safehouse_dir = None

def init():
    global configuration
    config_file = utils.find_file_in_parent_dirs('.shmake')
    if config_file:
        configuration = configurations.from_file(config_file)

init()



from ..exceptions import CommandError
from . import down, up


def run(args: list[str]):
    if len(args) < 1:
        raise CommandError(usage())
    command = args[0]
    if command == 'down':
        down.run()
    elif command == 'up':
        up.run(detach=True)
    else:
        raise CommandError(usage())
        

def usage() -> str:
    return "safehouse services <up|down>"



import json
from pathlib import Path

from safehouse.events.event_manager import EventManager
from safehouse import mode
from safehouse.services import local
from safehouse.types import JsonBlob
from . import exceptions


MODES = (
    'live',
    'local',
    'standalone',
)


class Project:
    def __init__(
        self,
        *,
        event_type_definitions: list[JsonBlob]=list(),
        filepath: Path,
        in_development: bool=False,
        mode: str='local',
        name: str,
        organization_name: str,
    ):
        self.event_type_definitions: list[JsonBlob] = event_type_definitions
        self._events_manager: EventManager | None = None
        self.filepath = filepath
        self.in_development = bool(in_development)
        self.mode = mode.lower()
        self.name = name.lower()
        self.organization_name = organization_name.lower()
        self.validate()

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{self.organization_name}.{self.name} ({self.filepath})"

    @property
    def events(self) -> EventManager:
        if self._events_manager is None:
            self._events_manager = EventManager(self.event_type_definitions)
            self._events_manager.register_event_types()
        return self._events_manager

    @property
    def is_local(self) -> bool:
        return self.mode == 'local'

    @property
    def json(self) -> JsonBlob:
        return {
            'events': self.event_type_definitions,
            'mode': self.mode,
            'name': self.name,
            'organization': self.organization_name,
        }

    @property
    def json_file(self) -> Path:
        return self.filepath / 'project.json'

    def save(self):
        with open(self.json_file, 'w') as f:
            json.dump(self.json, f, indent=4)

    @property
    def services_url(self) ->str:
        # use inversion of control here: services or actions (for standalone) interrogate for mode
        # and then "to the right thing"
        return mode.services_url()

    def validate(self):
        if self.in_development:
            assert self.filepath is not None
        assert self.name is not None
        assert self.organization_name is not None
        if self.mode not in MODES:
            raise Exception(f"unsupported mode: '{self.mode}'")
        #if self.is_local and not local.is_running():
        #    raise exceptions.ProjectError(f"{self} needs safehouse services to be running locally")


def from_safehouse_dir(safehouse_dir: Path, in_development: bool=False) -> Project:
    project_json_file = str(safehouse_dir / "project.json")
    with open(project_json_file, 'r', encoding='utf-8') as f:
        project_dict: JsonBlob = json.load(f)
        events = project_dict.get('events', {})
        mode = project_dict.get('mode', 'local')
        name = project_dict.get('name')
        org_name = project_dict.get('organization')
        return Project(
            event_type_definitions=events,
            filepath=safehouse_dir,
            in_development=in_development,
            mode=mode,
            name=name,
            organization_name=org_name,
        )


def from_settings(
        *,
        filepath: Path,
        mode: str,
        name: str,
        org_name: str,
) -> Project:
    return Project(
        filepath=filepath,
        mode=mode,
        name=name,
        organization_name=org_name,
    )
